#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import silver_platter   # noqa: F401
from silver_platter.debian import (
    _changelog_stats,
    )

from breezy.branch import Branch

import argparse

parser = argparse.ArgumentParser()
parser.add_argument('location', help='Branch location to check.', type=str,
                    default='.')
args = parser.parse_args()

branch = Branch.open(args.location)
print(_changelog_stats(branch, 200))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import subprocess

from .changer import (
    run_mutator,
    DebianChanger,
    ChangerResult,
    )
from breezy.trace import note


BRANCH_NAME = 'cme'


class CMEResult(object):

    def __init__(self):
        pass


class CMEChanger(DebianChanger):

    name = 'cme'

    def __init__(self, dry_run=False):
        self.dry_run = dry_run

    @classmethod
    def setup_parser(cls, parser):
        pass

    @classmethod
    def from_args(cls, args):
        return cls()

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(self, local_tree, subpath, update_changelog, reporter,
                     committer, base_proposal=None):
        base_revid = local_tree.last_revision()
        cwd = local_tree.abspath(subpath or '')
        subprocess.check_call(
            ['/usr/bin/cme', 'modify', 'dpkg', '-save'],
            cwd=cwd)
        local_tree.commit('Reformat for cme.')
        subprocess.check_call(
            ['/usr/bin/cme', 'fix', 'dpkg'], cwd=cwd)
        revid = local_tree.commit('Run cme.')
        branches = [('main', None, base_revid, revid)]
        tags = []
        return ChangerResult(
            description=None, mutator=None, branches=branches,
            tags=tags, proposed_commit_message='Run cme.',
            sufficient_for_proposal=True)

    def get_proposal_description(
            self, applied, description_format, existing_proposal):
        return 'Run cme.'

    def describe(self, result, publish_result):
        if publish_result.is_new:
            note('Proposed change from \'cme fix dpkg\': %s',
                 publish_result.proposal.url)
        else:
            note('No changes for package %s', result.package_name)

    @classmethod
    def describe_command(cls, command):
        return "Apply Configuration Model Editor (CME) fixes"


if __name__ == '__main__':
    import sys
    sys.exit(run_mutator(CMEChanger))



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from urllib.parse import urlparse

from . import (
    pick_additional_colocated_branches,
    add_changelog_entry,
    )
from .changer import (
    run_mutator,
    DebianChanger,
    ChangerError,
    ChangerResult,
    )
from ..proposal import push_changes
from breezy import osutils
from breezy.trace import note

from debmutate.control import ControlEditor
from debmutate.reformatting import GeneratedFile, FormattingUnpreservable


BRANCH_NAME = 'orphan'


def push_to_salsa(local_tree, user, name, dry_run=False):
    from breezy.branch import Branch
    from breezy.plugins.propose.gitlabs import GitLab
    salsa = GitLab.probe_from_url('https://salsa.debian.org/')
    # TODO(jelmer): Fork if the old branch was hosted on salsa
    if dry_run:
        note('Creating and pushing to salsa project %s/%s',
             user, name)
        return
    salsa.create_project('%s/%s' % (user, name))
    target_branch = Branch.open(
        'git+ssh://git@salsa.debian.org/%s/%s.git' % (user, name))
    additional_colocated_branches = pick_additional_colocated_branches(
        local_tree.branch)
    return push_changes(
        local_tree.branch, target_branch, hoster=salsa,
        additional_colocated_branches=additional_colocated_branches,
        dry_run=dry_run)


class OrphanResult(object):

    def __init__(self, package=None, old_vcs_url=None, new_vcs_url=None,
                 salsa_user=None):
        self.package = package
        self.old_vcs_url = old_vcs_url
        self.new_vcs_url = new_vcs_url
        self.pushed = False
        self.salsa_user = salsa_user


class OrphanChanger(DebianChanger):

    name = 'orphan'

    def __init__(self, update_vcs=True, salsa_push=True,
                 salsa_user='debian', dry_run=False):
        self.update_vcs = update_vcs
        self.salsa_push = salsa_push
        self.salsa_user = salsa_user
        self.dry_run = dry_run

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument(
            '--no-update-vcs', action='store_true',
            help='Do not move the VCS repository to the Debian team on Salsa.')
        parser.add_argument(
            '--salsa-user', type=str, default='debian',
            help='Salsa user to push repository to.')
        parser.add_argument(
            '--just-update-headers', action='store_true',
            help='Update the VCS-* headers, but don\'t actually '
            'clone the repository.')

    @classmethod
    def from_args(cls, args):
        return cls(
            update_vcs=not args.no_update_vcs,
            dry_run=args.dry_run,
            salsa_user=args.salsa_user,
            salsa_push=not args.just_update_headers)

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(self, local_tree, subpath, update_changelog,
                     reporter, committer, base_proposal=None):
        base_revid = local_tree.last_revision()
        control_path = local_tree.abspath(
                osutils.pathjoin(subpath, 'debian/control'))
        try:
            with ControlEditor(path=control_path) as editor:
                editor.source[
                    'Maintainer'] = 'Debian QA Group <packages@qa.debian.org>'
                try:
                    del editor.source['Uploaders']
                except KeyError:
                    pass

            result = OrphanResult()

            if self.update_vcs:
                with ControlEditor(path=control_path) as editor:
                    result.package_name = editor.source['Source']
                    result.old_vcs_url = editor.source.get('Vcs-Git')
                    editor.source['Vcs-Git'] = (
                        'https://salsa.debian.org/%s/%s.git' % (
                            self.salsa_user, result.package_name))
                    result.new_vcs_url = editor.source['Vcs-Git']
                    editor.source['Vcs-Browser'] = (
                        'https://salsa.debian.org/%s/%s' % (
                            self.salsa_user, result.package_name))
                    result.salsa_user = self.salsa_user
                if result.old_vcs_url == result.new_vcs_url:
                    result.old_vcs_url = result.new_vcs_url = None
            if update_changelog in (True, None):
                add_changelog_entry(
                    local_tree,
                    osutils.pathjoin(subpath, 'debian/changelog'),
                    ['QA Upload.', 'Move package to QA team.'])
            local_tree.commit(
                'Move package to QA team.', committer=committer,
                allow_pointless=False)
        except FormattingUnpreservable as e:
            raise ChangerError(
                'formatting-unpreservable',
                'unable to preserve formatting while editing %s' % e.path)
        except GeneratedFile as e:
            raise ChangerError(
                'generated-file',
                'unable to edit generated file: %r' % e)

        if self.update_vcs and self.salsa_push and result.new_vcs_url:
            push_to_salsa(
                local_tree, self.salsa_user, result.package_name,
                dry_run=self.dry_run)
            result.pushed = True
            reporter.report_metadata('old_vcs_url', result.old_vcs_url)
            reporter.report_metadata('new_vcs_url', result.new_vcs_url)
            reporter.report_metadata('pushed', result.pushed)

        branches = [
            ('main', None, base_revid,
             local_tree.last_revision())]

        tags = []

        return ChangerResult(
            description='Move package to QA team.',
            mutator=result, branches=branches, tags=tags,
            sufficient_for_proposal=True,
            proposed_commit_message=(
                'Set the package maintainer to the QA team.'))

    def get_proposal_description(
            self, applied, description_format, existing_proposal):
        return 'Set the package maintainer to the QA team.'

    def describe(self, result, publish_result):
        if publish_result.is_new:
            note('Proposed change of maintainer to QA team: %s',
                 publish_result.proposal.url)
        else:
            note('No changes for orphaned package %s', result.package_name)
        if result.pushed:
            note('Pushed new package to %s.', result.new_vcs_url)
        elif result.new_vcs_url:
            for line in move_instructions(
                    result.package_name, result.salsa_user, result.old_vcs_url,
                    result.new_vcs_url):
                note('%s', line)

    @classmethod
    def describe_command(cls, command):
        return "Mark as orphaned"


def move_instructions(package_name, salsa_user, old_vcs_url, new_vcs_url):
    yield 'Please move the repository from %s to %s.' % (
         old_vcs_url, new_vcs_url)
    if urlparse(old_vcs_url).hostname == 'salsa.debian.org':
        path = urlparse(old_vcs_url).path
        if path.endswith('.git'):
            path = path[:-4]
        yield 'If you have the salsa(1) tool installed, run: '
        yield ''
        yield '    salsa fork --group=%s %s' % (
             salsa_user, path)
    else:
        yield 'If you have the salsa(1) tool installed, run: '
        yield ''
        yield '    git clone %s %s' % (old_vcs_url, package_name)
        yield '    salsa --group=%s push_repo %s' % (salsa_user, package_name)


if __name__ == '__main__':
    import sys
    sys.exit(run_mutator(OrphanChanger))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import shutil
import socket
import subprocess
from typing import Callable, Tuple, Optional, List

from breezy import (
    config as _mod_config,
    errors,
    osutils,
    urlutils,
    )
try:
    from breezy.bzr import LineEndingError
except ImportError:  # brz < 3.1.1
    from breezy.errors import LineEndingError

from breezy.branch import (
    Branch,
    BranchWriteLockResult,
    )
from breezy.controldir import ControlDir, Prober
from breezy.git.remote import RemoteGitError
from breezy.revision import NULL_REVISION
from breezy.transport import Transport, get_transport
from breezy.workingtree import WorkingTree

from breezy.transport import UnusableRedirect


def create_temp_sprout(
        branch: Branch,
        additional_colocated_branches: Optional[List[str]] = None,
        dir: Optional[str] = None,
        path: Optional[str] = None) -> Tuple[WorkingTree, Callable[[], None]]:
    """Create a temporary sprout of a branch.

    This attempts to fetch the least amount of history as possible.
    """
    if path is None:
        td = osutils.mkdtemp(dir=dir)
    else:
        td = path
        os.mkdir(path)

    def destroy() -> None:
        shutil.rmtree(td)
    # Only use stacking if the remote repository supports chks because of
    # https://bugs.launchpad.net/bzr/+bug/375013
    use_stacking = (
        branch._format.supports_stacking() and
        branch.repository._format.supports_chks)
    try:
        # preserve whatever source format we have.
        to_dir = branch.controldir.sprout(
            td, None, create_tree_if_local=True,
            source_branch=branch,
            stacked=use_stacking)
        # TODO(jelmer): Fetch these during the initial clone
        for branch_name in set(additional_colocated_branches or []):
            try:
                add_branch = branch.controldir.open_branch(name=branch_name)
            except (errors.NotBranchError,
                    errors.NoColocatedBranchSupport):
                pass
            else:
                local_add_branch = to_dir.create_branch(name=branch_name)
                add_branch.push(local_add_branch)
                assert (add_branch.last_revision() ==
                        local_add_branch.last_revision())
        return to_dir.open_workingtree(), destroy
    except BaseException as e:
        destroy()
        raise e


class TemporarySprout(object):
    """Create a temporary sprout of a branch.

    This attempts to fetch the least amount of history as possible.
    """

    def __init__(self,
                 branch: Branch,
                 additional_colocated_branches: Optional[List[str]] = None,
                 dir: Optional[str] = None):
        self.branch = branch
        self.additional_colocated_branches = additional_colocated_branches
        self.dir = dir

    def __enter__(self) -> WorkingTree:
        tree, self._destroy = create_temp_sprout(
            self.branch,
            additional_colocated_branches=self.additional_colocated_branches,
            dir=self.dir)
        return tree

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._destroy()
        return False


class PreCheckFailed(Exception):
    """The post check failed."""


def run_pre_check(tree: WorkingTree, script: Optional[str]) -> None:
    """Run a script ahead of making any changes to a tree.

    Args:
      tree: The working tree to operate in
      script: Command to run
    Raises:
      PreCheckFailed: If the pre-check failed
    """
    if not script:
        return
    try:
        subprocess.check_call(script, shell=True, cwd=tree.basedir)
    except subprocess.CalledProcessError:
        raise PreCheckFailed()


class PostCheckFailed(Exception):
    """The post check failed."""


def run_post_check(tree: WorkingTree,
                   script: Optional[str],
                   since_revid: bytes) -> None:
    """Run a script after making any changes to a tree.

    Args:
      tree: The working tree to operate in
      script: Command to run
      since_revid: Revision id since which changes were made
    Raises:
      PostCheckFailed: If the pre-check failed
    """
    if not script:
        return
    try:
        subprocess.check_call(
            script, shell=True, cwd=tree.basedir,
            env={'SINCE_REVID': since_revid})
    except subprocess.CalledProcessError:
        raise PostCheckFailed()


class BranchUnavailable(Exception):
    """Opening branch failed."""

    def __init__(self, url: str, description: str):
        self.url = url
        self.description = description

    def __str__(self) -> str:
        return self.description


class BranchMissing(Exception):
    """Branch did not exist."""

    def __init__(self, url: str, description: str):
        self.url = url
        self.description = description

    def __str__(self) -> str:
        return self.description


class BranchUnsupported(Exception):
    """The branch uses a VCS or protocol that is unsupported."""

    def __init__(self, url: str, description: str):
        self.url = url
        self.description = description

    def __str__(self) -> str:
        return self.description


def _convert_exception(url: str, e: Exception) -> Optional[Exception]:
    if isinstance(e, socket.error):
        return BranchUnavailable(url, 'Socket error: %s' % e)
    if isinstance(e, errors.NotBranchError):
        return BranchMissing(url, 'Branch does not exist: %s' % e)
    if isinstance(e, errors.UnsupportedProtocol):
        return BranchUnsupported(url, str(e))
    if isinstance(e, errors.ConnectionError):
        return BranchUnavailable(url, str(e))
    if isinstance(e, errors.PermissionDenied):
        return BranchUnavailable(url, str(e))
    if isinstance(e,  errors.InvalidHttpResponse):
        return BranchUnavailable(url, str(e))
    if isinstance(e, errors.TransportError):
        return BranchUnavailable(url, str(e))
    if UnusableRedirect is not None and isinstance(e, UnusableRedirect):
        return BranchUnavailable(url, str(e))
    if isinstance(e, errors.UnsupportedFormatError):
        return BranchUnsupported(url, str(e))
    if isinstance(e, errors.UnknownFormatError):
        return BranchUnsupported(url, str(e))
    if isinstance(e, RemoteGitError):
        return BranchUnavailable(url, str(e))
    if isinstance(e, LineEndingError):
        return BranchUnavailable(url, str(e))
    return None


def open_branch(url: str,
                possible_transports: Optional[List[Transport]] = None,
                probers: Optional[List[Prober]] = None,
                name: str = None) -> Branch:
    """Open a branch by URL."""
    url, params = urlutils.split_segment_parameters(url)
    if name is None:
        try:
            name = urlutils.unquote(params['branch'])
        except KeyError:
            name = None
    try:
        transport = get_transport(
            url, possible_transports=possible_transports)
        dir = ControlDir.open_from_transport(transport, probers)
        return dir.open_branch(name=name)
    except Exception as e:
        converted = _convert_exception(url, e)
        if converted is not None:
            raise converted
        raise e


def open_branch_containing(
        url: str,
        possible_transports: Optional[List[Transport]] = None,
        probers: Optional[List[Prober]] = None) -> Tuple[Branch, str]:
    """Open a branch by URL."""
    try:
        transport = get_transport(
            url, possible_transports=possible_transports)
        dir, subpath = ControlDir.open_containing_from_transport(
            transport, probers)
        return dir.open_branch(), subpath
    except Exception as e:
        converted = _convert_exception(url, e)
        if converted is not None:
            raise converted
        raise e


try:
    from breezy.memorybranch import MemoryBranch
except ImportError:   # breezy < 3.1.1
    from breezy.lock import _RelockDebugMixin, LogicalLockResult

    class MemoryBranch(Branch, _RelockDebugMixin):  # type: ignore

        def __init__(self, repository, last_revision_info, tags):
            from breezy.tag import DisabledTags, MemoryTags
            self.repository = repository
            self._last_revision_info = last_revision_info
            self._revision_history_cache = None
            if tags is not None:
                self.tags = MemoryTags(tags)
            else:
                self.tags = DisabledTags(self)
            self._partial_revision_history_cache = []
            self._last_revision_info_cache = None
            self._revision_id_to_revno_cache = None
            self._partial_revision_id_to_revno_cache = {}
            self._partial_revision_history_cache = []
            self.base = 'memory://' + osutils.rand_chars(10)

        def get_config(self):
            return _mod_config.Config()

        def lock_read(self):
            self.repository.lock_read()
            return LogicalLockResult(self.unlock)

        def lock_write(self, token=None):
            self.repository.lock_write()
            return BranchWriteLockResult(self.unlock, None)

        def unlock(self):
            self.repository.unlock()

        def last_revision_info(self):
            return self._last_revision_info

        def _gen_revision_history(self):
            """Generate the revision history from last revision
            """
            last_revno, last_revision = self.last_revision_info()
            self._extend_partial_history()
            return list(reversed(self._partial_revision_history_cache))

        def get_rev_id(self, revno, history=None):
            """Find the revision id of the specified revno."""
            with self.lock_read():
                if revno == 0:
                    return NULL_REVISION
                last_revno, last_revid = self.last_revision_info()
                if revno == last_revno:
                    return last_revid
                if last_revno is None:
                    self._extend_partial_history()
                    return self._partial_revision_history_cache[
                            len(self._partial_revision_history_cache) - revno]
                else:
                    if revno <= 0 or revno > last_revno:
                        raise errors.NoSuchRevision(self, revno)
                    distance_from_last = last_revno - revno
                    if len(self._partial_revision_history_cache) <= \
                            distance_from_last:
                        self._extend_partial_history(distance_from_last)
                    return self._partial_revision_history_cache[
                        distance_from_last]


def full_branch_url(branch):
    """Get the full URL for a branch.

    Ideally this should just return Branch.user_url,
    but that currently exclude the branch name
    in some situations.
    """
    if branch.name is None:
        return branch.user_url
    url, params = urlutils.split_segment_parameters(branch.user_url)
    if branch.name != '':
        params['branch'] = urlutils.quote(branch.name, '')
    return urlutils.join_segment_parameters(url, params)



# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import sys

import breezy
from breezy.trace import note

from lintian_brush import (
    version_string as lintian_brush_version_string,
    )
from lintian_brush.debianize import (
    debianize,
    )
from lintian_brush.config import Config

import silver_platter

from . import (
    control_files_in_root,
    )
from .changer import (
    DebianChanger,
    ChangerResult,
    run_mutator,
    ChangerError,
    )


BRANCH_NAME = "debianize"


class DebianizeChanger(DebianChanger):

    name = 'debianize'

    def __init__(self, compat_release=None):
        self.compat_release = compat_release

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument(
            '--compat-release', type=str, help=argparse.SUPPRESS)

    @classmethod
    def from_args(cls, args):
        return cls(compat_release=args.compat_release)

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(self, local_tree, subpath, update_changelog,
                     reporter, committer, base_proposal=None):
        base_revid = local_tree.last_revision()

        reporter.report_metadata('versions', {
            'lintian-brush': lintian_brush_version_string,
            'silver-platter': silver_platter.version_string,
            'breezy': breezy.version_string,
        })

        import distro_info
        debian_info = distro_info.DebianDistroInfo()

        compat_release = self.compat_release
        try:
            cfg = Config.from_workingtree(local_tree, subpath)
        except FileNotFoundError:
            pass
        else:
            compat_release = cfg.compat_release()
            if compat_release:
                compat_release = debian_info.codename(
                    compat_release, default=compat_release)
        if compat_release is None:
            compat_release = debian_info.stable()

        with local_tree.lock_write():
            if control_files_in_root(local_tree, subpath):
                raise ChangerError(
                    'control-files-in-root',
                    'control files live in root rather than debian/ '
                    '(LarstIQ mode)')

            debianize(
                local_tree, subpath=subpath,
                compat_release=self.compat_release)

        branches = [
            ('main', None, base_revid,
             local_tree.last_revision())]

        return ChangerResult(
            description='Debianized package.',
            mutator=None,
            branches=branches, tags=[],
            value=None,
            sufficient_for_proposal=True)

    def get_proposal_description(
            self, applied, description_format, existing_proposal):
        return "Debianize package."

    def describe(self, applied, publish_result):
        note('Created Debian package.')


if __name__ == '__main__':
    sys.exit(run_mutator(DebianizeChanger))



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy.tests import TestCaseWithTransport


from ..debian.upstream import (
    update_packaging,
    )


class UpdatePackagingTests(TestCaseWithTransport):

    def test_autogen_sh(self):
        t = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
lintian-brush (0.37) UNRELEASED; urgency=medium

  * Add various more aliases for salsa team names.

 -- Jelmer Vernooĳ <jelmer@debian.org>  Fri, 18 Oct 2019 17:34:35 +0000
"""),
            ('debian/rules', """\
%:
\tdh %
""")])
        t.add(['debian', 'debian/changelog', 'debian/rules'])
        oldrev = t.commit('Initial')

        self.build_tree_contents([
            ('autogen.sh', '#!/bin/sh\n')])
        t.add(['autogen.sh'])
        t.commit('Add autogen')

        self.addCleanup(t.lock_write().unlock)
        update_packaging(
            t, t.branch.repository.revision_tree(oldrev),
            committer="Jelmer Vernooĳ <jelmer@debian.org>")

        self.assertFileEqual("""\
lintian-brush (0.37) UNRELEASED; urgency=medium

  * Add various more aliases for salsa team names.
  * Invoke autogen.sh from dh_autoreconf.

 -- Jelmer Vernooĳ <jelmer@debian.org>  Fri, 18 Oct 2019 17:34:35 +0000
""", 'debian/changelog')
        self.assertFileEqual("""\
%:
\tdh %

override_dh_autoreconf:
\tdh_autoreconf ./autogen.sh
""", 'debian/rules')



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for uploading packages."""

import silver_platter   # noqa: F401

import datetime
from email.utils import parseaddr
import os
import subprocess
import sys
import tempfile
from typing import List

from debmutate.changelog import (
    ChangelogEditor,
    changeblock_ensure_first_line,
    )
from debmutate.control import ControlEditor

from breezy import gpg
from breezy.errors import NoSuchTag
from breezy.commit import NullCommitReporter
from breezy.plugins.debian.builder import BuildFailedError
from breezy.plugins.debian.cmds import _build_helper
from breezy.plugins.debian.import_dsc import (
    DistributionBranch,
    )
from breezy.plugins.debian.release import (
    release,
    )
from breezy.plugins.debian.util import (
    changelog_find_previous_upload,
    dput_changes,
    find_changelog,
    debsign,
    )
from breezy.trace import note, show_error, warning

from debian.changelog import get_maintainer

from . import (
    apt_get_source_package,
    source_package_vcs,
    split_vcs_url,
    Workspace,
    DEFAULT_BUILDER,
    select_probers,
    )
from ..utils import (
    open_branch,
    BranchUnavailable,
    BranchMissing,
    BranchUnsupported,
    )


class NoUnuploadedChanges(Exception):
    """Indicates there are no unuploaded changes for a package."""

    def __init__(self, archive_version):
        self.archive_version = archive_version
        super(NoUnuploadedChanges, self).__init__(
            "nothing to upload, latest version is in archive: %s" %
            archive_version)


class NoUnreleasedChanges(Exception):
    """Indicates there are no unreleased changes for a package."""

    def __init__(self, version):
        self.version = version
        super(NoUnreleasedChanges, self).__init__(
            "nothing to upload, latest version in vcs is not unreleased: %s" %
            version)


class RecentCommits(Exception):
    """Indicates there are too recent commits for a package."""

    def __init__(self, commit_age, min_commit_age):
        self.commit_age = commit_age
        self.min_commit_age = min_commit_age
        super(RecentCommits, self).__init__(
            "Last commit is only %d days old (< %d)" % (
                self.commit_age, self.min_commit_age))


class LastReleaseRevisionNotFound(Exception):
    """The revision for the last uploaded release can't be found."""

    def __init__(self, package, version):
        self.package = package
        self.version = version
        super(LastReleaseRevisionNotFound, self).__init__(
            "Unable to find revision matching version %r for %s" % (
                version, package))


def check_revision(rev, min_commit_age):
    """Check whether a revision can be included in an upload.

    Args:
      rev: revision to check
      min_commit_age: Minimum age for revisions
    Raises:
      RecentCommits: When there are commits younger than min_commit_age
    """
    # TODO(jelmer): deal with timezone
    if min_commit_age is not None:
        commit_time = datetime.datetime.fromtimestamp(rev.timestamp)
        time_delta = datetime.datetime.now() - commit_time
        if time_delta.days < min_commit_age:
            raise RecentCommits(time_delta.days, min_commit_age)
    # TODO(jelmer): Allow tag to prevent automatic uploads


def find_last_release_revid(branch, version):
    db = DistributionBranch(branch, None)
    return db.revid_of_version(version)


def get_maintainer_keys(context):
    for key in context.keylist(
            source='/usr/share/keyrings/debian-keyring.gpg'):
        yield key.fpr
        for subkey in key.subkeys:
            yield subkey.keyid


def prepare_upload_package(
        local_tree, subpath, pkg, last_uploaded_version, builder,
        gpg_strategy=None, min_commit_age=None):
    if local_tree.has_filename(os.path.join(subpath, 'debian/gbp.conf')):
        subprocess.check_call(
            ['gbp', 'dch', '--ignore-branch'],
            cwd=local_tree.abspath('.'))
    cl, top_level = find_changelog(local_tree, merge=False, max_blocks=None)
    if cl.version == last_uploaded_version:
        raise NoUnuploadedChanges(cl.version)
    previous_version_in_branch = changelog_find_previous_upload(cl)
    if last_uploaded_version > previous_version_in_branch:
        raise Exception(
            "last uploaded version more recent than previous "
            "version in branch: %r > %r" % (
                last_uploaded_version, previous_version_in_branch))

    note("Checking revisions since %s" % last_uploaded_version)
    with local_tree.lock_read():
        try:
            last_release_revid = find_last_release_revid(
                    local_tree.branch, last_uploaded_version)
        except NoSuchTag:
            raise LastReleaseRevisionNotFound(pkg, last_uploaded_version)
        graph = local_tree.branch.repository.get_graph()
        revids = list(graph.iter_lefthand_ancestry(
            local_tree.branch.last_revision(), [last_release_revid]))
        if not revids:
            note("No pending changes")
            return
        if gpg_strategy:
            note('Verifying GPG signatures...')
            count, result, all_verifiables = gpg.bulk_verify_signatures(
                    local_tree.branch.repository, revids,
                    gpg_strategy)
            for revid, code, key in result:
                if code != gpg.SIGNATURE_VALID:
                    raise Exception(
                        "No valid GPG signature on %r: %d" %
                        (revid, code))
        for revid, rev in local_tree.branch.repository.iter_revisions(
                revids):
            check_revision(rev, min_commit_age)

        if cl.distributions != "UNRELEASED":
            raise NoUnreleasedChanges(cl.version)
    qa_upload = False
    team_upload = False
    control_path = local_tree.abspath(os.path.join(subpath, 'debian/control'))
    with ControlEditor(control_path) as e:
        maintainer = parseaddr(e.source['Maintainer'])
        if maintainer[1] == 'packages@qa.debian.org':
            qa_upload = True
        # TODO(jelmer): Check whether this is a team upload
        # TODO(jelmer): determine whether this is a NMU upload
    if qa_upload or team_upload:
        changelog_path = local_tree.abspath(
            os.path.join(subpath, 'debian/changelog'))
        with ChangelogEditor(changelog_path) as e:
            if qa_upload:
                changeblock_ensure_first_line(e[0], 'QA upload.')
            elif team_upload:
                changeblock_ensure_first_line(e[0], 'Team upload.')
        local_tree.commit(
            specific_files=[os.path.join(subpath, 'debian/changelog')],
            message='Mention QA Upload.',
            allow_pointless=False,
            reporter=NullCommitReporter())
    tag_name = release(local_tree, subpath)
    target_dir = tempfile.mkdtemp()
    builder = builder.replace("${LAST_VERSION}", last_uploaded_version)
    target_changes = _build_helper(
        local_tree, subpath, local_tree.branch, target_dir, builder=builder)
    debsign(target_changes)
    return target_changes, tag_name


def select_apt_packages(package_names, maintainer):
    packages = []
    import apt_pkg
    apt_pkg.init()
    sources = apt_pkg.SourceRecords()
    while sources.step():
        if maintainer:
            fullname, email = parseaddr(sources.maintainer)
            if email not in maintainer:
                continue

        if package_names and sources.package not in package_names:
            continue

        packages.append(sources.package)

    return packages


def select_vcswatch_packages(packages: List[str], maintainer: List[str]):
    import psycopg2
    conn = psycopg2.connect(
        database="udd",
        user="udd-mirror",
        password="udd-mirror",
        host="udd-mirror.debian.net")
    cursor = conn.cursor()
    args = []
    query = """\
    SELECT sources.source, vcswatch.url
    FROM vcswatch JOIN sources ON sources.source = vcswatch.source
    WHERE
     vcswatch.status IN ('COMMITS', 'NEW') AND
     sources.release = 'sid'
"""
    if maintainer:
        query += " AND sources.maintainer_email IN (%s)"
        args.append(tuple(maintainer))
    if packages:
        query += " AND sources.source IN (%s)"
        args.append(tuple(packages))

    cursor.execute(query, tuple(args))

    packages = []
    for package, vcs_url in cursor.fetchall():
        packages.append(package)
    return packages


def main(argv):
    import argparse
    parser = argparse.ArgumentParser(prog='upload-pending-commits')
    parser.add_argument("packages", nargs='*')
    parser.add_argument(
        '--dry-run', action='store_true',
        help='Dry run changes.')
    parser.add_argument(
        '--acceptable-keys',
        help='List of acceptable GPG keys',
        action='append', default=[], type=str)
    parser.add_argument(
        '--gpg-verification',
        help='Verify GPG signatures on commits', action='store_true')
    parser.add_argument(
        '--min-commit-age',
        help='Minimum age of the last commit, in days',
        type=int, default=0)
    parser.add_argument(
        '--diff',
        action='store_true',
        help='Show diff.')
    parser.add_argument(
        '--builder',
        type=str,
        help='Build command',
        default=(DEFAULT_BUILDER + ' --source --source-only-changes '
                 '--debbuildopt=-v${LAST_VERSION}'))
    parser.add_argument(
        '--maintainer',
        type=str,
        action='append',
        help='Select all packages maintainer by specified maintainer.')
    parser.add_argument(
        '--vcswatch',
        action='store_true',
        default=False,
        help='Use vcswatch to determine what packages need uploading.')
    parser.add_argument(
        '--autopkgtest-only',
        action='store_true',
        help='Only process packages with autopkgtests.')

    args = parser.parse_args(argv)
    ret = 0

    if not args.packages and not args.maintainer:
        (name, email) = get_maintainer()
        if email:
            note('Processing packages maintained by %s', email)
            args.maintainer = [email]
        else:
            parser.print_usage()
            sys.exit(1)

    if args.vcswatch:
        packages = select_vcswatch_packages(
            args.packages, args.maintainer)
    else:
        note('Use --vcswatch to only process packages for which '
             'vcswatch found pending commits.')
        if args.maintainer:
            packages = select_apt_packages(
                args.packages, args.maintainer)
        else:
            packages = args.packages

    if not packages:
        note('No packages found.')
        parser.print_usage()
        sys.exit(1)

    # TODO(jelmer): Sort packages by last commit date; least recently changed
    # commits are more likely to be successful.

    if len(packages) > 1:
        note('Uploading packages: %s', ', '.join(packages))

    for package in packages:
        note('Processing %s', package)
        # Can't use open_packaging_branch here, since we want to use pkg_source
        # later on.
        if '/' not in package:
            pkg_source = apt_get_source_package(package)
            try:
                vcs_type, vcs_url = source_package_vcs(pkg_source)
            except KeyError:
                note('%s: no declared vcs location, skipping',
                     pkg_source['Package'])
                ret = 1
                continue
            source_name = pkg_source['Package']
            source_version = pkg_source['Version']
            has_testsuite = 'Testsuite' in pkg_source
        else:
            vcs_url = package
            vcs_type = None
            source_name = None
            source_version = None
            has_testsuite = None
        (location, branch_name, subpath) = split_vcs_url(vcs_url)
        if subpath is None:
            subpath = ''
        probers = select_probers(vcs_type)
        try:
            main_branch = open_branch(
                location, probers=probers, name=branch_name)
        except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
            show_error('%s: %s', vcs_url, e)
            ret = 1
            continue
        with Workspace(main_branch) as ws:
            if source_name is None:
                with ControlEditor(
                        ws.local_tree.abspath(
                            os.path.join(subpath, 'debian/control'))
                        ) as ce:
                    source_name = ce.source['Source']
                with ChangelogEditor(
                        ws.local_tree.abspath(
                            os.path.join(subpath, 'debian/changelog'))
                        ) as cle:
                    source_version = cle[0].version
                has_testsuite = 'Testsuite' in ce.source
            if (args.autopkgtest_only and
                    not has_testsuite and
                    not ws.local_tree.has_filename(
                        os.path.join(subpath, 'debian/tests/control'))):
                note('%s: Skipping, package has no autopkgtest.',
                     source_name)
                continue
            branch_config = ws.local_tree.branch.get_config_stack()
            if args.gpg_verification:
                gpg_strategy = gpg.GPGStrategy(branch_config)
                if args.acceptable_keys:
                    acceptable_keys = args.acceptable_keys
                else:
                    acceptable_keys = list(get_maintainer_keys(
                        gpg_strategy.context))
                gpg_strategy.set_acceptable_keys(','.join(acceptable_keys))
            else:
                gpg_strategy = None

            try:
                target_changes, tag_name = prepare_upload_package(
                    ws.local_tree, subpath,
                    source_name, source_version,
                    builder=args.builder, gpg_strategy=gpg_strategy,
                    min_commit_age=args.min_commit_age)
            except BuildFailedError as e:
                warning(
                    '%s: package failed to build: %s',
                    source_name, e)
                ret = 1
                continue
            except LastReleaseRevisionNotFound as e:
                warning(
                    '%s: Unable to find revision matching last release '
                    '%s, skipping.', source_name, e.version)
                ret = 1
                continue
            except RecentCommits as e:
                note('%s: Recent commits (%d days), skipping.',
                     source_name, e.commit_age)
                continue
            except NoUnuploadedChanges:
                note('%s: No unuploaded changes, skipping.',
                     source_name)
                continue
            except NoUnreleasedChanges:
                note('%s: No unreleased changes, skipping.',
                     source_name)
                continue

            tags = []
            if tag_name is not None:
                note('Pushing tag %s', tag_name)
                tags.append(tag_name)
            ws.push(dry_run=args.dry_run, tags=tags)
            if not args.dry_run:
                dput_changes(target_changes)
            if args.diff:
                sys.stdout.flush()
                ws.show_diff(sys.stdout.buffer)
                sys.stdout.buffer.flush()

    return ret


if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for scrubbing obsolete settings."""

import argparse

from debmutate.reformatting import GeneratedFile, FormattingUnpreservable

import silver_platter  # noqa: F401

from lintian_brush.config import Config

from .changer import (
    DebianChanger,
    ChangerError,
    ChangerResult,
    run_mutator,
    )

from breezy.trace import note

BRANCH_NAME = 'scrub-obsolete'
DEFAULT_VALUE_MULTIARCH_HINT = 30


def calculate_value(result):
    value = DEFAULT_VALUE_MULTIARCH_HINT
    for para, changes in result.control_removed:
        for field, packages in changes:
            value += len(packages) * 2
    for path, removed in result.maintscript_removed:
        value += len(removed)
    return value


class ScrubObsoleteChanger(DebianChanger):

    name: str = 'scrub-obsolete'

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            '--allow-reformatting', default=None, action='store_true',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--upgrade-release', metavar='UPGRADE-RELEASE',
            help='Release to allow upgrading from.', default='oldstable')

    @classmethod
    def from_args(cls, args):
        import distro_info
        debian_info = distro_info.DebianDistroInfo()
        upgrade_release = debian_info.codename(args.upgrade_release)
        return cls(
            allow_reformatting=args.allow_reformatting,
            upgrade_release=upgrade_release)

    def __init__(self, upgrade_release, allow_reformatting=None):
        self.allow_reformatting = allow_reformatting
        self.upgrade_release = upgrade_release

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(self, local_tree, subpath, update_changelog,
                     reporter, committer, base_proposal=None):
        from lintian_brush.scrub_obsolete import scrub_obsolete
        base_revid = local_tree.last_revision()
        allow_reformatting = self.allow_reformatting
        try:
            cfg = Config.from_workingtree(local_tree, subpath)
        except FileNotFoundError:
            pass
        else:
            if allow_reformatting is None:
                allow_reformatting = cfg.allow_reformatting()
            if update_changelog is None:
                update_changelog = cfg.update_changelog()

        try:
            result = scrub_obsolete(
                local_tree, subpath, self.upgrade_release,
                update_changelog=update_changelog)
        except FormattingUnpreservable as e:
            raise ChangerError(
                'formatting-unpreservable',
                'unable to preserve formatting while editing %s' % e.path)
        except GeneratedFile as e:
            raise ChangerError(
                'generated-file',
                'unable to edit generated file: %r' % e)

        if not result:
            raise ChangerError('nothing-to-do', 'no obsolete constraints')

        branches = [
            ('main', None, base_revid,
             local_tree.last_revision())]

        tags = []

        return ChangerResult(
            description="Scrub obsolete settings.", mutator=result,
            branches=branches, tags=tags,
            value=calculate_value(result),
            sufficient_for_proposal=True,
            proposed_commit_message='Scrub obsolete settings.')

    def get_proposal_description(
            self, result, description_format, existing_proposal):
        ret = [
            'Remove constraints unnecessary since %s.' % self.upgrade_release,
            ''] + ['* ' + line for line in result.itemized()]
        return ''.join(ret)

    def describe(self, applied, publish_result):
        note('Scrub obsolete settings.')
        for line in applied.itemized():
            note('* %s', line)

    @classmethod
    def describe_command(cls, command):
        return "Remove obsolete dependencies"


if __name__ == '__main__':
    import sys
    sys.exit(run_mutator(ScrubObsoleteChanger))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import sys
from typing import List, Set

from debian.changelog import ChangelogCreateError

import breezy
from breezy.errors import BzrError
from breezy.trace import note

from lintian_brush import (
    available_lintian_fixers,
    run_lintian_fixers,
    DEFAULT_MINIMUM_CERTAINTY,
    SUPPORTED_CERTAINTIES,
    NotDebianPackage,
    version_string as lintian_brush_version_string,
    )
from lintian_brush.config import Config

import silver_platter

from . import (
    control_files_in_root,
    )
from .changer import (
    DebianChanger,
    ChangerResult,
    run_mutator,
    ChangerError,
    )

__all__ = [
    'available_lintian_fixers',
    'calculate_value',
    ]


DEFAULT_ADDON_FIXERS = [
    'debian-changelog-line-too-long',
    'file-contains-trailing-whitespace',
    'out-of-date-standards-version',
    'package-uses-old-debhelper-compat-version',
    'public-upstream-key-not-minimal',
    'no-dh-sequencer',
    ]

DEFAULT_VALUE_LINTIAN_BRUSH_ADDON_ONLY = 10
DEFAULT_VALUE_LINTIAN_BRUSH = 50
# Base these scores on the importance as set in Debian?
LINTIAN_BRUSH_TAG_VALUES = {
    'file-contains-trailing-whitespace': 0,
    }
LINTIAN_BRUSH_TAG_DEFAULT_VALUE = 5


BRANCH_NAME = "lintian-fixes"


class UnknownFixer(BzrError):
    """The specified fixer is unknown."""

    _fmt = "No such fixer: %s."

    def __init__(self, fixer):
        super(UnknownFixer, self).__init__(fixer=fixer)


def calculate_value(tags: Set[str]) -> int:
    if not (set(tags) - set(DEFAULT_ADDON_FIXERS)):
        value = DEFAULT_VALUE_LINTIAN_BRUSH_ADDON_ONLY
    else:
        value = DEFAULT_VALUE_LINTIAN_BRUSH
    for tag in tags:
        value += LINTIAN_BRUSH_TAG_VALUES.get(
            tag, LINTIAN_BRUSH_TAG_DEFAULT_VALUE)
    return value


def parse_mp_description(description: str) -> List[str]:
    """Parse a merge proposal description.

    Args:
      description: The description to parse
    Returns:
      list of one-line descriptions of changes
    """
    existing_lines = description.splitlines()
    if len(existing_lines) == 1:
        return existing_lines
    else:
        return [line[2:].rstrip('\n')
                for line in existing_lines if line.startswith('* ')]


def create_mp_description(description_format: str, lines: List[str]) -> str:
    """Create a merge proposal description.

    Args:
      lines: List of one-line descriptions of fixes
    Returns:
      A string with a merge proposal description
    """
    if len(lines) > 1:
        mp_description = ["Fix some issues reported by lintian\n"]
        for line in lines:
            line = "* %s\n" % line
            if line not in mp_description:
                mp_description.append(line)
    else:
        mp_description = [lines[0]]
    return ''.join(mp_description)


def applied_entry_as_line(description_format, fixed_lintian_tags, line):
    if not fixed_lintian_tags:
        return line
    if description_format == 'markdown':
        return '%s (%s)' % (line, ', '.join(
            ['[%s](https://lintian.debian.org/tags/%s.html)' % (tag, tag)
             for tag in fixed_lintian_tags]))
    return '%s (%s)' % (line, ', '.join(fixed_lintian_tags))


def update_proposal_description(
        description_format, existing_proposal, applied):
    if existing_proposal:
        existing_description = existing_proposal.get_description()
        existing_lines = parse_mp_description(existing_description)
    else:
        existing_lines = []
    return create_mp_description(
        description_format, existing_lines +
        [applied_entry_as_line(description_format, r.fixed_lintian_tags, l)
         for r, l in applied])


def update_proposal_commit_message(existing_proposal, applied):
    existing_commit_message = getattr(
        existing_proposal, 'get_commit_message', lambda: None)()
    if existing_commit_message and not existing_commit_message.startswith(
            'Fix lintian issues: '):
        # The commit message is something we haven't set - let's leave it
        # alone.
        return
    if existing_commit_message:
        existing_applied = existing_commit_message.split(':', 1)[1]
    else:
        existing_applied = []
    return "Fix lintian issues: " + (
        ', '.join(sorted(existing_applied + [l for r, l in applied])))


def has_nontrivial_changes(applied, propose_addon_only: Set[str]) -> bool:
    tags = set()
    for result, unused_summary in applied:
        tags.update(result.fixed_lintian_tags)
    # Is there enough to create a new merge proposal?
    return bool(tags - set(propose_addon_only))


def get_fixers(available_fixers, names=None, tags=None, exclude=None):
    """Get the set of fixers to try.

    Args:
      available_fixers: Dictionary mapping fixer names to objects
      names: Optional set of fixers to restrict to
      tags: Optional set of tags to restrict to
      exclude: Optional set of fixers to exclude
    Returns:
      List of fixer objects
    """
    if exclude is None:
        exclude = set()
    by_tag = {}
    by_name = {}
    for fixer in available_fixers:
        for tag in fixer.lintian_tags:
            by_tag[tag] = fixer
        by_name[fixer.name] = fixer

    # If it's unknown which fixers are relevant, just try all of them.
    if names:
        try:
            fixers = [by_name[name] for name in names]
        except KeyError as e:
            raise UnknownFixer(e.args[0])
    elif tags:
        fixers = [by_tag[tag] for tag in tags]
    else:
        fixers = list(by_name.values())

    if exclude:
        fixers = [fixer for fixer in fixers if fixer.name not in exclude]
    return fixers


class LintianBrushChanger(DebianChanger):

    name = 'lintian-brush'

    def __init__(
            self, names=None, exclude=None, propose_addon_only=None,
            compat_release=None, allow_reformatting=None,
            minimum_certainty=None, tags=None,
            opinionated=False, trust_package=False, diligence=0):
        self.fixers = get_fixers(
            available_lintian_fixers(), names=names,
            tags=tags, exclude=exclude)
        self.propose_addon_only = propose_addon_only or []
        self.compat_release = compat_release
        self.allow_reformatting = allow_reformatting
        self.minimum_certainty = minimum_certainty
        self.opinionated = opinionated
        self.trust_package = trust_package
        self.diligence = diligence

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument(
            "--fixers",
            help="Fixers to run.", type=str, action='append')
        parser.add_argument(
            '--exclude',
            help='Fixers to exclude.', type=str, action='append')
        parser.add_argument(
            '--propose-addon-only',
            help='Fixers that should be considered add-on-only.',
            type=str, action='append',
            default=DEFAULT_ADDON_FIXERS)
        parser.add_argument(
            '--compat-release', type=str, default=None,
            help='Oldest Debian release to be compatible with.')
        parser.add_argument(
            '--allow-reformatting', default=None, action='store_true',
            help='Whether to allow reformatting.')
        parser.add_argument(
            '--minimum-certainty',
            type=str,
            choices=SUPPORTED_CERTAINTIES,
            default=None,
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--opinionated', action='store_true',
            help='Make opinionated changes')
        parser.add_argument(
            '--diligence', type=int, default=10,
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--trust-package', action='store_true',
            help='Trust package.')
        parser.add_argument("tags", nargs='*')

    @classmethod
    def from_args(cls, args):
        return cls(names=args.fixers, exclude=args.exclude,
                   propose_addon_only=args.propose_addon_only,
                   compat_release=args.compat_release,
                   allow_reformatting=args.allow_reformatting,
                   minimum_certainty=args.minimum_certainty,
                   tags=args.tags, opinionated=args.opinionated,
                   diligence=args.diligence,
                   trust_package=args.trust_package)

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(self, local_tree, subpath, update_changelog,
                     reporter, committer, base_proposal=None):
        base_revid = local_tree.last_revision()

        reporter.report_metadata('versions', {
            'lintian-brush': lintian_brush_version_string,
            'silver-platter': silver_platter.version_string,
            'breezy': breezy.version_string,
        })

        import distro_info
        debian_info = distro_info.DebianDistroInfo()

        compat_release = self.compat_release
        allow_reformatting = self.allow_reformatting
        minimum_certainty = None
        try:
            cfg = Config.from_workingtree(local_tree, subpath)
        except FileNotFoundError:
            pass
        else:
            compat_release = cfg.compat_release()
            if compat_release:
                compat_release = debian_info.codename(
                    compat_release, default=compat_release)
            allow_reformatting = cfg.allow_reformatting()
            minimum_certainty = cfg.minimum_certainty()
        if compat_release is None:
            compat_release = debian_info.stable()
        if allow_reformatting is None:
            allow_reformatting = False
        if minimum_certainty is None:
            minimum_certainty = DEFAULT_MINIMUM_CERTAINTY

        with local_tree.lock_write():
            if control_files_in_root(local_tree, subpath):
                raise ChangerError(
                    'control-files-in-root',
                    'control files live in root rather than debian/ '
                    '(LarstIQ mode)')

            try:
                overall_result = run_lintian_fixers(
                        local_tree, self.fixers,
                        committer=committer,
                        update_changelog=update_changelog,
                        compat_release=compat_release,
                        allow_reformatting=allow_reformatting,
                        minimum_certainty=minimum_certainty,
                        subpath=subpath, diligence=self.diligence,
                        opinionated=self.opinionated,
                        trust_package=self.trust_package)
            except NotDebianPackage:
                raise ChangerError(
                    'not-debian-package', 'Not a Debian package')
            except ChangelogCreateError as e:
                raise ChangerError(
                    'changelog-create-error',
                    'Error creating changelog entry: %s' % e)

        applied = []
        base_applied = reporter.get_base_metadata('applied', [])
        if base_applied:
            applied.extend(base_applied)
        for result, summary in overall_result.success:
            applied.append({
                'summary': summary,
                'description': result.description,
                'fixed_lintian_tags': result.fixed_lintian_tags,
                'revision_id': result.revision_id.decode('utf-8'),
                'certainty': result.certainty})
        reporter.report_metadata('applied', applied)

        if overall_result.failed_fixers:
            for fixer_name, failure in overall_result.failed_fixers.items():
                note('Fixer %r failed to run:', fixer_name)
                sys.stderr.write(str(failure))
        reporter.report_metadata(
            'failed', {
                name: str(e)
                for (name, e) in overall_result.failed_fixers.items()})

        if not overall_result.success:
            raise ChangerError('nothing-to-do', 'no fixers to apply')

        fixed_lintian_tags = set()
        for result, summary in overall_result.success:
            fixed_lintian_tags.update(result.fixed_lintian_tags)

        add_on_only = not has_nontrivial_changes(
            overall_result.success, self.propose_addon_only)

        if not reporter.get_base_metadata('add_on_only', False):
            add_on_only = False

        if not add_on_only:
            if overall_result.success:
                note('only add-on fixers found')
            sufficient_for_proposal = False
            reporter.report_metadata('add_on_only', True)
        else:
            sufficient_for_proposal = True
            reporter.report_metadata('add_on_only', False)

        branches = [
            ('main', None, base_revid,
             local_tree.last_revision())]

        return ChangerResult(
            description='Applied fixes for %r' % fixed_lintian_tags,
            mutator=overall_result.success,
            branches=branches, tags=[],
            value=calculate_value(fixed_lintian_tags),
            sufficient_for_proposal=sufficient_for_proposal,
            proposed_commit_message=update_proposal_commit_message(
                base_proposal, overall_result.success))

    def get_proposal_description(
            self, applied, description_format, existing_proposal):
        return update_proposal_description(
            description_format, existing_proposal, applied)

    def describe(self, applied, publish_result):
        tags = set()
        for brush_result, unused_summary in applied:
            tags.update(brush_result.fixed_lintian_tags)
        if publish_result.is_new:
            note('Proposed fixes %r: %s', tags, publish_result.proposal.url)
        elif tags:
            note('Updated proposal %s with fixes %r',
                 publish_result.proposal.url, tags)
        else:
            note('No new fixes for proposal %s', publish_result.proposal.url)


if __name__ == '__main__':
    sys.exit(run_mutator(LintianBrushChanger))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from typing import Optional, Dict, List, Callable

import silver_platter  # noqa: F401

import argparse
import sys

from .changer import (
    setup_parser_common,
    run_single_changer,
    changer_subcommands,
    changer_subcommand,
    )

from . import uploader as debian_uploader


def run_changer_subcommand(name, changer_cls, argv):
    parser = argparse.ArgumentParser(prog='debian-svp %s URL|package' % name)
    setup_parser_common(parser)
    parser.add_argument('package', type=str, nargs='?')
    changer_cls.setup_parser(parser)
    args = parser.parse_args(argv)
    if args.package is None:
        parser.print_usage()
        return 1
    changer = changer_cls.from_args(args)
    return run_single_changer(changer, args)


def main(argv: Optional[List[str]] = None) -> Optional[int]:
    import breezy
    breezy.initialize()

    from ..__main__ import subcommands as main_subcommands

    subcommands: Dict[
            str, Callable[[List[str]], Optional[int]]] = {
        'upload-pending': debian_uploader.main,
        }

    parser = argparse.ArgumentParser(prog='debian-svp', add_help=False)
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s ' + silver_platter.version_string)
    parser.add_argument(
        '--help', action='store_true',
        help='show this help message and exit')

    subcommands.update(main_subcommands.items())

    # We have a debian-specific run command
    del subcommands['run']

    parser.add_argument(
        'subcommand', type=str,
        choices=list(subcommands.keys()) + changer_subcommands())
    args, rest = parser.parse_known_args()
    if args.help:
        if args.subcommand is None:
            parser.print_help()
            parser.exit()
        else:
            rest.append('--help')

    if args.subcommand is None:
        parser.print_usage()
        return 1
    if args.subcommand in subcommands:
        return subcommands[args.subcommand](rest)
    try:
        subcmd = changer_subcommand(args.subcommand)
    except KeyError:
        pass
    else:
        return run_changer_subcommand(args.subcommand, subcmd, rest)
    parser.print_usage()
    return 1


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from distro_info import DebianDistroInfo

import os
import re
import tempfile

from . import (
    DEFAULT_BUILDER,
    )
from .changer import (
    run_mutator,
    DebianChanger,
    ChangerResult,
    )
from breezy.plugins.debian.cmds import _build_helper
from breezy.plugins.debian.util import (
    dput_changes,
    debsign,
    )
from breezy.trace import note

from debian.changelog import format_date, get_maintainer
from debmutate.changelog import ChangelogEditor, changeblock_add_line


# See https://backports.debian.org/Contribute/


class BackportResult(object):

    def __init__(self, target_release):
        self.target_release = target_release


def backport_suffix(release):
    distro_info = DebianDistroInfo()
    version = distro_info.version(release)
    return 'bpo%s' % version


def backport_distribution(release):
    distro_info = DebianDistroInfo()
    if distro_info.codename('stable') == release:
        return '%s-backports' % release
    elif distro_info.codename('oldstable') == release:
        return '%s-backports-sloppy' % release
    else:
        raise Exception('unable to determine target suite for %s' % release)


def create_bpo_version(orig_version, bpo_suffix):
    m = re.fullmatch(r'(.*)\~' + bpo_suffix + r'\+([0-9]+)', str(orig_version))
    if m:
        base = m.group(1)
        buildno = int(m.group(2)) + 1
    else:
        base = str(orig_version)
        buildno = 1
    return '%s~%s+%d' % (base, bpo_suffix, buildno)


def backport_package(local_tree, subpath, target_release, author=None):
    changes = []
    # TODO(jelmer): Check that package has a high enough popcon count,
    # and warn otherwise?
    # TODO(jelmer): Iterate Build-Depends and verify that depends are
    # satisfied by target_distribution
    # TODO(jelmer): Update Vcs-Git/Vcs-Browser header?
    target_distribution = backport_distribution(target_release)
    version_suffix = backport_suffix(target_release)
    note('Using target distribution %s, version suffix %s',
         target_distribution, version_suffix)
    clp = local_tree.abspath(os.path.join(subpath, 'debian/changelog'))

    if author is None:
        author = '%s <%s>' % get_maintainer()

    with ChangelogEditor(clp) as cl:
        # TODO(jelmer): If there was an existing backport, use that version
        since_version = cl[0].version
        cl.new_block(
            package=cl[0].package,
            distributions=target_distribution,
            urgency='low',
            author=author,
            date=format_date(),
            version=create_bpo_version(since_version, version_suffix))
        block = cl[0]
        changeblock_add_line(
            block,
            ['Backport to %s.' % target_release] +
            [' +' + line for line in changes])

    return since_version


class BackportChanger(DebianChanger):

    name = 'backport'

    def __init__(self, dry_run=False, target_release=None, builder=None):
        self.dry_run = dry_run
        self.target_release = target_release
        self.builder = builder

    @classmethod
    def setup_parser(cls, parser):
        distro_info = DebianDistroInfo()
        parser.add_argument(
            '--target-release', type=str,
            help='Target release', default=distro_info.stable())
        parser.add_argument(
            '--dry-run', action='store_true',
            help='Do a dry run.')
        parser.add_argument(
            '--builder',
            type=str,
            help='Build command',
            default=(DEFAULT_BUILDER + ' --source --source-only-changes '
                     '--debbuildopt=-v${LAST_VERSION}'))

    @classmethod
    def from_args(cls, args):
        return cls(target_release=args.target_release, dry_run=args.dry_run,
                   builder=args.builder)

    def suggest_branch_name(self):
        return backport_distribution(self.target_release)

    def make_changes(self, local_tree, subpath, update_changelog, reporter,
                     committer, base_proposal=None):

        base_revision = local_tree.last_revision()

        since_version = backport_package(
            local_tree, subpath, self.target_release, author=committer)

        with tempfile.TemporaryDirectory() as td:
            builder = self.builder.replace(
                "${LAST_VERSION}", str(since_version))
            target_changes = _build_helper(
                local_tree, subpath, local_tree.branch, td, builder=builder)
            debsign(target_changes)

            if not self.dry_run:
                dput_changes(target_changes)

        branches = [
            ('main', None, base_revision,
             local_tree.last_revision())]

        # TODO(jelmer): Add debian/... tag
        tags = []

        return ChangerResult(
            description=None, mutator=None, branches=branches,
            tags=tags,
            proposed_commit_message='Backport to %s.' % self.target_release,
            sufficient_for_proposal=True)

    def get_proposal_description(
            self, applied, description_format, existing_proposal):
        return 'Backport to %s.' % self.target_release

    def describe(self, result, publish_result):
        if publish_result.is_new:
            note('Proposed backportg to %s: %s',
                 result.target_release,
                 publish_result.proposal.url)
        else:
            note('No changes for package %s', result.package_name)


if __name__ == '__main__':
    import sys
    sys.exit(run_mutator(BackportChanger))



#!/usr/bin/python
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from typing import Optional, Callable, List, Union, Dict, BinaryIO, Any, Tuple

from breezy.branch import Branch
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.diff import show_diff_trees
from breezy.errors import (
    DivergedBranches,
    NotBranchError,
    NoColocatedBranchSupport,
    )
from breezy.propose import (
    get_hoster,
    Hoster,
    MergeProposal,
    )

from breezy.trace import note

from .publish import (
    merge_conflicts,
    propose_changes,
    push_changes,
    push_derived_changes,
    publish_changes as _publish_changes,
    PublishResult,
    )
from .utils import (
    create_temp_sprout,
    full_branch_url,
    )


__all__ = [
    'Workspace',
    ]


class Workspace(object):
    """Workspace for creating changes to a branch.

    main_branch: The upstream branch
    resume_branch: Optional in-progress branch that we previously made changes
        on, and should ideally continue from.
    cached_branch: Branch to copy revisions from, if possible.
    local_tree: The tree the user can work in
    """

    _destroy: Optional[Callable[[], None]]
    local_tree: WorkingTree
    main_branch_revid: Optional[bytes]

    def __init__(self, main_branch: Branch,
                 resume_branch: Optional[Branch] = None,
                 cached_branch: Optional[Branch] = None,
                 additional_colocated_branches: Optional[List[str]] = None,
                 dir: Optional[str] = None,
                 path: Optional[str] = None) -> None:
        self.main_branch = main_branch
        self.main_branch_revid = None
        self.cached_branch = cached_branch
        self.resume_branch = resume_branch
        self.additional_colocated_branches = (
            additional_colocated_branches or [])
        self._destroy = None
        self._dir = dir
        self._path = path

    def __str__(self):
        if self._path is None:
            return "Workspace for %s" % full_branch_url(self.main_branch)
        else:
            return "Workspace for %s at %s" % (
                full_branch_url(self.main_branch), self._path)

    def __repr__(self):
        return (
            "%s(%r, resume_branch=%r, cached_branch=%r, "
            "additional_colocated_branches=%r, dir=%r, path=%r)" % (
                type(self).__name__, self.main_branch, self.resume_branch,
                self.cached_branch, self.additional_colocated_branches,
                self._dir, self._path))

    def __enter__(self) -> Any:
        self.local_tree, self._destroy = create_temp_sprout(
            self.cached_branch or self.resume_branch or self.main_branch,
            self.additional_colocated_branches,
            dir=self._dir, path=self._path)
        self.main_branch_revid = self.main_branch.last_revision()
        self.refreshed = False
        with self.local_tree.branch.lock_write():
            if self.cached_branch:
                self.local_tree.pull(
                    self.resume_branch or self.main_branch, overwrite=True)
            if self.resume_branch:
                try:
                    self.local_tree.pull(self.main_branch, overwrite=False)
                except DivergedBranches:
                    pass
                for branch_name in self.additional_colocated_branches:
                    try:
                        remote_colo_branch = (
                            self.main_branch.controldir.open_branch(
                                name=branch_name))
                    except (NotBranchError,
                            NoColocatedBranchSupport):
                        continue
                    self.local_tree.branch.controldir.push_branch(
                        name=branch_name, source=remote_colo_branch,
                        overwrite=True)
                if merge_conflicts(self.main_branch, self.local_tree.branch):
                    note('restarting branch')
                    self.local_tree.update(revision=self.main_branch_revid)
                    self.local_tree.branch.generate_revision_history(
                        self.main_branch_revid)
                    self.resume_branch = None
                    self.refreshed = True
            self.orig_revid = self.local_tree.last_revision()
        return self

    def defer_destroy(self) -> Optional[Callable[[], None]]:
        ret = self._destroy
        self._destroy = None
        return ret

    def changes_since_main(self) -> bool:
        return self.local_tree.branch.last_revision() != self.main_branch_revid

    def changes_since_resume(self) -> bool:
        return self.orig_revid != self.local_tree.branch.last_revision()

    def push(self, hoster: Optional[Hoster] = None, dry_run: bool = False,
             tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
             stop_revision: Optional[bytes] = None) -> None:
        if hoster is None:
            hoster = get_hoster(self.main_branch)
        return push_changes(
            self.local_tree.branch, self.main_branch, hoster=hoster,
            additional_colocated_branches=self.additional_colocated_branches,
            dry_run=dry_run, tags=tags, stop_revision=stop_revision)

    def propose(self, name: str, description: str,
                hoster: Optional[Hoster] = None,
                existing_proposal: Optional[MergeProposal] = None,
                overwrite_existing: Optional[bool] = None,
                labels: Optional[List[str]] = None,
                dry_run: bool = False,
                commit_message: Optional[str] = None,
                reviewers: Optional[List[str]] = None,
                tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
                owner: Optional[str] = None,
                allow_collaboration: bool = False,
                stop_revision: Optional[bytes] = None) -> MergeProposal:
        if hoster is None:
            hoster = get_hoster(self.main_branch)
        return propose_changes(
            self.local_tree.branch, self.main_branch, hoster=hoster, name=name,
            mp_description=description, resume_branch=self.resume_branch,
            resume_proposal=existing_proposal,
            overwrite_existing=(overwrite_existing or False), labels=labels,
            dry_run=dry_run, commit_message=commit_message,
            reviewers=reviewers, owner=owner,
            additional_colocated_branches=self.additional_colocated_branches,
            tags=tags, allow_collaboration=allow_collaboration,
            stop_revision=stop_revision)

    def push_derived(self,
                     name: str, hoster: Optional[Hoster] = None,
                     overwrite_existing: Optional[bool] = False,
                     owner: Optional[str] = None,
                     tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
                     stop_revision: Optional[bytes] = None
                     ) -> Tuple[Branch, str]:
        """Push a derived branch.

        Args:
          name: Branch name
          hoster: Optional hoster to use
          overwrite_existing: Whether to overwrite an existing branch
          tags: Tags list to push
          owner: Owner name
        Returns:
          tuple with remote_branch and public_branch_url
        """
        if hoster is None:
            hoster = get_hoster(self.main_branch)
        return push_derived_changes(
            self.local_tree.branch,
            self.main_branch, hoster, name,
            overwrite_existing=overwrite_existing,
            owner=owner, tags=tags, stop_revision=stop_revision)

    def publish_changes(self, *args, **kwargs) -> PublishResult:
        """Publish a set of changes.
        """
        return _publish_changes(
            self.local_tree.branch, self.main_branch, self.resume_branch,
            *args, **kwargs)

    def orig_tree(self) -> Tree:
        return self.local_tree.branch.repository.revision_tree(self.orig_revid)

    def show_diff(self, outf: BinaryIO,
                  old_label: str = 'old/', new_label: str = 'new/') -> None:
        orig_tree = self.orig_tree()
        show_diff_trees(
            orig_tree, self.local_tree.basis_tree(), outf,
            old_label=old_label, new_label=new_label)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._destroy:
            self._destroy()
            self._destroy = None
        return False



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import contextlib
import json
import os
import subprocess
import tempfile

from urllib.request import urlopen

from debian.changelog import Changelog

from . import NoAptSources
from .changer import (
    run_mutator,
    DebianChanger,
    ChangerResult,
    ChangerError,
    )
from breezy.trace import note
from breezy.plugins.debian.upstream import PackageVersionNotPresent


BRANCH_NAME = 'missing-commits'


class AptSourceError(Exception):
    """An error occured while running 'apt source'."""

    def __init__(self, reason):
        self.reason = reason


def select_vcswatch_packages():
    import psycopg2
    conn = psycopg2.connect(
        database="udd",
        user="udd-mirror",
        password="udd-mirror",
        host="udd-mirror.debian.net")
    cursor = conn.cursor()
    args = []
    query = """\
    SELECT sources.source, vcswatch.url
    FROM vcswatch JOIN sources ON sources.source = vcswatch.source
    WHERE
     vcswatch.status IN ('OLD', 'UNREL') AND
     sources.release = 'sid'
"""
    cursor.execute(query, tuple(args))
    packages = []
    for package, vcs_url in cursor.fetchall():
        packages.append(package)
    return packages


def download_snapshot(package, version, output_dir, no_preparation=False):
    note('Downloading %s %s', package, version)
    srcfiles_url = ('https://snapshot.debian.org/mr/package/%s/%s/'
                    'srcfiles?fileinfo=1' % (package, version))
    files = {}
    for hsh, entries in json.load(urlopen(srcfiles_url))['fileinfo'].items():
        for entry in entries:
            files[entry['name']] = hsh
    for filename, hsh in files.items():
        local_path = os.path.join(output_dir, os.path.basename(filename))
        with open(local_path, 'wb') as f:
            url = 'https://snapshot.debian.org/file/%s' % hsh
            with urlopen(url) as g:
                f.write(g.read())
    args = []
    if no_preparation:
        args.append('--no-preparation')
    subprocess.check_call(
        ['dpkg-source'] + args + ['-x', '%s_%s.dsc' % (package, version)],
        cwd=output_dir)


class NoMissingVersions(Exception):

    def __init__(self, vcs_version, archive_version):
        self.vcs_version = vcs_version
        self.archive_version = archive_version
        super(NoMissingVersions, self).__init__(
            'No missing versions after all. Archive has %s, VCS has %s' % (
                archive_version, vcs_version))


class TreeVersionNotInArchiveChangelog(Exception):

    def __init__(self, tree_version):
        self.tree_version = tree_version
        super(TreeVersionNotInArchiveChangelog, self).__init__(
            'tree version %s does not appear in archive changelog' %
            tree_version)


class TreeUpstreamVersionMissing(Exception):

    def __init__(self, upstream_version):
        self.upstream_version = upstream_version
        super(TreeUpstreamVersionMissing, self).__init__(
            'unable to find upstream version %r' % upstream_version)


def retrieve_source(package_name, target):
    try:
        subprocess.run(
            ['apt', 'source', package_name], cwd=target,
            check=True,
            stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        stderr = e.stderr.splitlines()
        if stderr[-1] == (
                b'E: You must put some \'source\' URIs in your '
                b'sources.list'):
            raise NoAptSources()
        CS = b"\x1b[1;31mE: \x1b[0m"
        CE = b"\x1b[0m"
        if stderr[-1] == (
                CS +
                b"You must put some 'deb-src' URIs in your sources.list" +
                CE):
            raise NoAptSources()
        if stderr[-1].startswith(b'E: '):
            raise AptSourceError(stderr[-1][3:].decode())
        if stderr[-1].startswith(CS):
            raise AptSourceError(stderr[-1][len(CS):-len(CE)])
        raise AptSourceError(
            [line.decode('utf-8', 'surrogateescape') for line in stderr])


def import_uncommitted(tree, subpath):
    from breezy.plugins.debian.import_dsc import (
        DistributionBranch,
        DistributionBranchSet,
        )
    cl_path = os.path.join(subpath, 'debian/changelog')
    with tree.get_file(cl_path) as f:
        tree_cl = Changelog(f)
        package_name = tree_cl.package
    with contextlib.ExitStack() as es:
        archive_source = es.enter_context(tempfile.TemporaryDirectory())
        try:
            retrieve_source(package_name, archive_source)
        except AptSourceError as e:
            if isinstance(e.reason, list):
                reason = e.reason[-1]
            else:
                reason = e.reason
            raise ChangerError('apt-source-error', reason)
        except NoAptSources:
            raise ChangerError(
                'no-apt-sources',
                'No sources configured in /etc/apt/sources.list')
        [subdir] = [
            e.path for e in os.scandir(archive_source) if e.is_dir()]
        with open(os.path.join(subdir, 'debian', 'changelog'), 'r') as f:
            archive_cl = Changelog(f)
        missing_versions = []
        for block in archive_cl:
            if block.version == tree_cl.version:
                break
            missing_versions.append(block.version)
        else:
            raise TreeVersionNotInArchiveChangelog(tree_cl.version)
        if len(missing_versions) == 0:
            raise NoMissingVersions(tree_cl.version, archive_cl.version)
        note('Missing versions: %s', ', '.join(map(str, missing_versions)))
        ret = []
        dbs = DistributionBranchSet()
        db = DistributionBranch(tree.branch, tree.branch, tree=tree)
        dbs.add_branch(db)
        if tree_cl.version.debian_revision:
            note('Extracting upstream version %s.',
                 tree_cl.version.upstream_version)
            upstream_dir = es.enter_context(tempfile.TemporaryDirectory())
            try:
                upstream_tips = db.pristine_upstream_source\
                    .version_as_revisions(
                        tree_cl.package,
                        tree_cl.version.upstream_version)
            except PackageVersionNotPresent:
                raise TreeUpstreamVersionMissing(
                    tree_cl.version.upstream_version)
            db.extract_upstream_tree(upstream_tips, upstream_dir)
        no_preparation = not tree.has_filename('.pc/applied-patches')
        version_path = {}
        for version in missing_versions:
            output_dir = es.enter_context(tempfile.TemporaryDirectory())
            download_snapshot(
                package_name, version, output_dir,
                no_preparation=no_preparation)
            version_path[version] = output_dir
        for version in reversed(missing_versions):
            note('Importing %s', version)
            dsc_path = os.path.join(
                version_path[version],
                '%s_%s.dsc' % (package_name, version))
            tag_name = db.import_package(dsc_path)
            revision = db.version_as_revisions(version)
            ret.append((tag_name, version, revision))
    return ret


class UncommittedChanger(DebianChanger):

    name = 'import-upload'

    @classmethod
    def setup_parser(cls, parser):
        pass

    @classmethod
    def from_args(cls, args):
        return cls()

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(self, local_tree, subpath, update_changelog, reporter,
                     committer, base_proposal=None):
        base_revid = local_tree.last_revision()
        try:
            ret = import_uncommitted(local_tree, subpath)
        except TreeUpstreamVersionMissing as e:
            raise ChangerError('tree-upstream-version-missing', str(e))
        except TreeVersionNotInArchiveChangelog as e:
            raise ChangerError(
                'tree-version-not-in-archive-changelog', str(e))
        except NoMissingVersions as e:
            raise ChangerError('nothing-to-do', str(e))
        tags = [(None, tag_name, revid) for (tag_name, version, revid) in ret]
        # TODO(jelmer): Include upstream tags
        proposed_commit_message = "Import missing uploads: %s." % (
            ', '.join([str(v) for t, v in ret]))
        reporter.report_metadata('tags', [
            (tag_name, str(version)) for (tag_name, version, revid) in ret])

        branches = [
            ('main', None, base_revid,
             local_tree.last_revision())]

        # TODO(jelmer): Include branches for upstream/pristine-tar

        return ChangerResult(
            description='Import archive changes missing from the VCS.',
            branches=branches, mutator=ret, tags=tags,
            sufficient_for_proposal=True,
            proposed_commit_message=proposed_commit_message)

    def get_proposal_description(
            self, applied, description_format, existing_proposal):
        return "Import missing uploads: %s." % (
            ', '.join([str(v) for t, v in applied]))

    def describe(self, applied, publish_result):
        if publish_result.is_new:
            note('Proposed import of versions %s: %s',
                 ', '.join([str(v) for t, v in applied]),
                 publish_result.proposal.url)
        elif applied:
            note('Updated proposal %s with versions %s.',
                 publish_result.proposal.url,
                 ', '.join([str(v) for t, v in applied]))
        else:
            note('No new versions imported for proposal %s',
                 publish_result.proposal.url)

    @classmethod
    def describe_command(cls, command):
        return "Import archive changes missing from VCS"


if __name__ == '__main__':
    import sys
    sys.exit(run_mutator(UncommittedChanger))



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__all__ = ['iter_conflicted']

import argparse
from functools import partial
import pkg_resources
import sys
from typing import Any, List, Optional, Dict, Iterable, Tuple, Type

from breezy import version_info as breezy_version_info
from breezy.branch import Branch
from breezy.propose import Hoster, MergeProposal
from breezy.trace import note, warning, show_error
from breezy.transport import Transport
from breezy.workingtree import WorkingTree

from . import (
    open_packaging_branch,
    guess_update_changelog,
    NoSuchPackage,
    NoAptSources,
    DEFAULT_BUILDER,
    )
from ..proposal import (
    HosterLoginRequired,
    UnsupportedHoster,
    NoSuchProject,
    enable_tag_pushing,
    find_existing_proposed,
    get_hoster,
    iter_conflicted,
    )

from ..publish import (
    PublishResult,
    SUPPORTED_MODES,
    )
from ..utils import (
    BranchMissing,
    BranchUnavailable,
    BranchUnsupported,
    run_pre_check,
    run_post_check,
    PostCheckFailed,
    full_branch_url,
    )


def get_package(package: str, branch_name: str,
                overwrite_unrelated: bool = False,
                refresh: bool = False,
                possible_transports: Optional[List[Transport]] = None,
                possible_hosters: Optional[List[Hoster]] = None,
                owner: Optional[str] = None
                ) -> Tuple[str, Branch, str, Optional[Branch],
                           Optional[Hoster], Optional[MergeProposal],
                           Optional[bool]]:
    main_branch, subpath = open_packaging_branch(
        package, possible_transports=possible_transports)

    overwrite: Optional[bool] = False

    try:
        hoster = get_hoster(main_branch, possible_hosters=possible_hosters)
    except UnsupportedHoster:
        # We can't figure out what branch to resume from when there's no
        # hoster that can tell us.
        resume_branch = None
        existing_proposal = None
        hoster = None
    else:
        (resume_branch, overwrite, existing_proposal) = (
            find_existing_proposed(
                main_branch, hoster, branch_name, owner=owner,
                overwrite_unrelated=overwrite_unrelated))
    if refresh:
        overwrite = True
        resume_branch = None

    return (
        package, main_branch, subpath, resume_branch, hoster,
        existing_proposal, overwrite)


def iter_packages(packages: Iterable[str], branch_name: str,
                  overwrite_unrelated: bool = False,
                  refresh: bool = False,
                  derived_owner: Optional[str] = None):
    """Iterate over relevant branches for a set of packages.

    Args:
      packages: Iterable over package names (or packaging URLs)
      branch_name: Branch name to look for
      overwrite_unrelated: Allow overwriting unrelated changes
      refresh: Whether to refresh existing merge proposals
    Returns:
      iterator over
        (package name, main branch object, subpath, branch to resume (if any),
         hoster (None if the hoster is not supported),
         existing_proposal, whether to overwrite the branch)
    """
    possible_transports: List[Transport] = []
    possible_hosters: List[Hoster] = []

    for pkg in packages:
        note('Processing: %s', pkg)

        (pkg, main_branch, subpath, resume_branch, hoster, existing_proposal,
         overwrite) = get_package(
                pkg, branch_name, overwrite_unrelated=overwrite_unrelated,
                refresh=refresh, possible_transports=possible_transports,
                possible_hosters=possible_hosters, owner=derived_owner)

        yield (pkg, main_branch, subpath, resume_branch, hoster,
               existing_proposal, overwrite)


class ChangerReporter(object):

    def report_context(self, context):
        raise NotImplementedError(self.report_context)

    def report_metadata(self, key, value):
        raise NotImplementedError(self.report_metadata)

    def get_base_metadata(self, key, default_value=None):
        raise NotImplementedError(self.get_base_metadata)


class ChangerError(Exception):

    def __init__(self, category: str, summary: str,
                 original: Optional[Exception] = None):
        self.category = category
        self.summary = summary
        self.original = original


class ChangerResult(object):

    def __init__(self, description: Optional[str], mutator: Any,
                 branches: Optional[List[Tuple[str, str, bytes, bytes]]] = [],
                 tags: Optional[Dict[str, bytes]] = None,
                 value: Optional[int] = None,
                 proposed_commit_message: Optional[str] = None,
                 title: Optional[str] = None,
                 labels: Optional[List[str]] = None,
                 sufficient_for_proposal: bool = True):
        self.description = description
        self.mutator = mutator
        self.branches = branches or []
        self.tags = tags or {}
        self.value = value
        self.proposed_commit_message = proposed_commit_message
        self.title = title
        self.labels = labels
        self.sufficient_for_proposal = sufficient_for_proposal

    def show_diff(self, repository, outf, role='main',
                  old_label: str = 'old/', new_label: str = 'new/') -> None:
        from breezy.diff import show_diff_trees
        for (brole, name, base_revision, revision) in self.branches:
            if role == brole:
                break
        else:
            raise KeyError
        old_tree = repository.revision_tree(base_revision)
        new_tree = repository.revision_tree(revision)
        show_diff_trees(
            old_tree, new_tree, outf,
            old_label=old_label, new_label=new_label)


def setup_parser_common(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--dry-run",
        help="Create branches but don't push or propose anything.",
        action="store_true", default=False)
    parser.add_argument(
        '--build-verify',
        help='Build package to verify it.',
        dest='build_verify',
        action='store_true')
    parser.add_argument(
        '--pre-check',
        help='Command to run to check whether to process package.',
        type=str)
    parser.add_argument(
        '--post-check',
        help='Command to run to check package before pushing.',
        type=str)
    parser.add_argument(
        '--builder', default=DEFAULT_BUILDER, type=str,
        help='Build command to use when verifying build.')
    parser.add_argument(
        '--refresh',
        help='Discard old branch and apply fixers from scratch.',
        action='store_true')
    parser.add_argument(
        '--committer',
        help='Committer identity',
        type=str)
    parser.add_argument(
        '--mode',
        help='Mode for pushing', choices=SUPPORTED_MODES,
        default="propose", type=str)
    parser.add_argument(
        '--no-update-changelog', action="store_false", default=None,
        dest="update_changelog", help="do not update the changelog")
    parser.add_argument(
        '--update-changelog', action="store_true", dest="update_changelog",
        help="force updating of the changelog", default=None)
    parser.add_argument(
        '--diff', action="store_true",
        help="Output diff of created merge proposal.")
    parser.add_argument(
        '--build-target-dir', type=str,
        help=("Store built Debian files in specified directory "
              "(with --build-verify)"))
    parser.add_argument(
        '--overwrite', action='store_true',
        help='Overwrite existing branches.')
    parser.add_argument(
        '--name', type=str,
        help='Proposed branch name', default=None)
    parser.add_argument(
        '--derived-owner', type=str, default=None,
        help='Owner for derived branches.')
    parser.add_argument(
        '--label', type=str,
        help='Label to attach', action="append", default=[])


class DebianChanger(object):
    """A class which can make and explain changes to a Debian package in VCS.
    """

    name: str

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser) -> None:
        raise NotImplementedError(cls.setup_parser)

    @classmethod
    def from_args(cls, args: List[str]) -> 'DebianChanger':
        raise NotImplementedError(cls.from_args)

    def suggest_branch_name(self) -> str:
        raise NotImplementedError(self.suggest_branch_name)

    def make_changes(self,
                     local_tree: WorkingTree,
                     subpath: str,
                     update_changelog: bool,
                     reporter: ChangerReporter,
                     committer: Optional[str],
                     base_proposal: Optional[MergeProposal] = None,
                     ) -> ChangerResult:
        raise NotImplementedError(self.make_changes)

    def get_proposal_description(
            self, applied: Any,
            description_format: str,
            existing_proposal: MergeProposal) -> str:
        raise NotImplementedError(self.get_proposal_description)

    def describe(self, applied: Any, publish_result: PublishResult) -> None:
        raise NotImplementedError(self.describe)

    @classmethod
    def describe_command(cls, command):
        return cls.name


class DummyChangerReporter(ChangerReporter):

    def report_context(self, context):
        pass

    def report_metadata(self, key, value):
        pass

    def get_base_metadata(self, key, default_value=None):
        return None


def _run_single_changer(
        changer: DebianChanger,
        pkg: str,
        main_branch: Branch,
        subpath: str,
        resume_branch: Optional[Branch],
        hoster: Optional[Hoster],
        existing_proposal: Optional[MergeProposal],
        overwrite: Optional[bool], mode: str, branch_name: str,
        diff: bool = False,
        committer: Optional[str] = None, build_verify: bool = False,
        pre_check: Optional[str] = None, post_check: Optional[str] = None,
        builder: str = DEFAULT_BUILDER, dry_run: bool = False,
        update_changelog: Optional[bool] = None,
        label: Optional[List[str]] = None,
        derived_owner: Optional[str] = None,
        build_target_dir: Optional[str] = None) -> Optional[bool]:
    from breezy import errors
    from . import (
        BuildFailedError,
        MissingUpstreamTarball,
        Workspace,
        )

    if hoster is None and mode == 'attempt-push':
        warning('Unsupported hoster; will attempt to push to %s',
                full_branch_url(main_branch))
        mode = 'push'
    with Workspace(main_branch, resume_branch=resume_branch) as ws, \
            ws.local_tree.lock_write():
        if ws.refreshed:
            overwrite = True
        run_pre_check(ws.local_tree, pre_check)
        if update_changelog is None:
            dch_guess = guess_update_changelog(ws.local_tree)
            if dch_guess:
                note(dch_guess[1])
                update_changelog = dch_guess[0]
            else:
                # Assume yes.
                update_changelog = True
        try:
            changer_result = changer.make_changes(
                ws.local_tree, subpath=subpath,
                update_changelog=update_changelog,
                committer=committer, reporter=DummyChangerReporter())
        except ChangerError as e:
            show_error(e.summary)
            return False

        if not ws.changes_since_main():
            if existing_proposal:
                note('%s: nothing left to do. Closing proposal.', pkg)
                existing_proposal.close()
            else:
                note('%s: nothing to do', pkg)
            return None

        try:
            run_post_check(ws.local_tree, post_check, ws.orig_revid)
        except PostCheckFailed as e:
            note('%s: %s', pkg, e)
            return False
        if build_verify:
            try:
                ws.build(builder=builder, result_dir=build_target_dir)
            except BuildFailedError:
                note('%s: build failed', pkg)
                return False
            except MissingUpstreamTarball:
                note('%s: unable to find upstream source', pkg)
                return False

        enable_tag_pushing(ws.local_tree.branch)

        kwargs: Dict[str, Any] = {}
        if breezy_version_info >= (3, 1):
            kwargs['tags'] = changer_result.tags

        try:
            publish_result = ws.publish_changes(
                mode, branch_name,
                get_proposal_description=partial(
                    changer.get_proposal_description, changer_result.mutator),
                get_proposal_commit_message=(
                    lambda oldmp: changer_result.proposed_commit_message),
                dry_run=dry_run, hoster=hoster,
                allow_create_proposal=changer_result.sufficient_for_proposal,
                overwrite_existing=overwrite,
                existing_proposal=existing_proposal,
                derived_owner=derived_owner, labels=label,
                **kwargs)
        except UnsupportedHoster as e:
            show_error(
                '%s: No known supported hoster for %s. Run \'svp login\'?',
                pkg, full_branch_url(e.branch))
            return False
        except NoSuchProject as e:
            note('%s: project %s was not found', pkg, e.project)
            return False
        except errors.PermissionDenied as e:
            note('%s: %s', pkg, e)
            return False
        except errors.DivergedBranches:
            note('%s: a branch exists. Use --overwrite to discard it.',
                 pkg)
            return False
        except HosterLoginRequired as e:
            show_error(
                'Credentials for hosting site at %r missing. '
                'Run \'svp login\'?', e.hoster.base_url)
            return False

        if publish_result.proposal:
            changer.describe(changer_result.mutator, publish_result)
        if diff:
            for entry in changer_result.branches:
                role = entry[0]
                if len(changer_result.branches) > 1:
                    sys.stdout.write('%s\n' % role)
                    sys.stdout.write(('-' * len(role)) + '\n')
                sys.stdout.flush()
                changer_result.show_diff(
                    ws.local_tree.branch.repository, sys.stdout.buffer,
                    role=role)
                if len(changer_result.branches) > 1:
                    sys.stdout.write('\n')

        return True


def run_single_changer(
        changer: DebianChanger, args: argparse.Namespace) -> int:
    import silver_platter   # noqa: F401

    if args.name:
        branch_name = args.name
    else:
        branch_name = changer.suggest_branch_name()

    try:
        (pkg, main_branch, subpath, resume_branch, hoster, existing_proposal,
         overwrite) = get_package(
                args.package, branch_name, overwrite_unrelated=args.overwrite,
                refresh=args.refresh, owner=args.derived_owner)
    except NoSuchPackage:
        note('%s: no such package', args.package)
        return 1
    except NoSuchProject as e:
        note('%s: unable to find project: %s', args.package, e.project)
        return 1
    except (BranchMissing, BranchUnavailable, BranchUnsupported) as e:
        note('%s: ignoring: %s', args.package, e)
        return 1
    except NoAptSources:
        note('%s: no apt sources configured, unable to get package metadata.',
             args.package)
        return 1

    if _run_single_changer(
            changer, pkg, main_branch, subpath, resume_branch, hoster,
            existing_proposal, overwrite, args.mode, branch_name,
            diff=args.diff, committer=args.committer,
            build_verify=args.build_verify,
            pre_check=args.pre_check, builder=args.builder,
            post_check=args.post_check, dry_run=args.dry_run,
            update_changelog=args.update_changelog,
            label=args.label, derived_owner=args.derived_owner,
            build_target_dir=args.build_target_dir) is False:
        return 1
    else:
        return 0


BUILTIN_ENTRYPOINTS = [
    pkg_resources.EntryPoint(
        'run', 'silver_platter.debian.run', attrs=('ScriptChanger', )),
    pkg_resources.EntryPoint(
        'lintian-brush', 'silver_platter.debian.lintian',
        attrs=('LintianBrushChanger', )),
    pkg_resources.EntryPoint(
        'tidy', 'silver_platter.debian.tidy',
        attrs=('TidyChanger', )),
    pkg_resources.EntryPoint(
        'new-upstream', 'silver_platter.debian.upstream',
        attrs=('NewUpstreamChanger', )),
    pkg_resources.EntryPoint(
        'cme', 'silver_platter.debian.cme',
        attrs=('CMEChanger', )),
    pkg_resources.EntryPoint(
        'apply-multiarch-hints',
        'silver_platter.debian.multiarch', attrs=('MultiArchHintsChanger', )),
    pkg_resources.EntryPoint(
        'rules-requires-root',
        'silver_platter.debian.rrr', attrs=('RulesRequiresRootChanger', )),
    pkg_resources.EntryPoint(
        'orphan',
        'silver_platter.debian.orphan', attrs=('OrphanChanger', )),
    pkg_resources.EntryPoint(
        'import-upload',
        'silver_platter.debian.uncommitted', attrs=('UncommittedChanger', )),
    pkg_resources.EntryPoint(
        'scrub-obsolete',
        'silver_platter.debian.scrub_obsolete',
        attrs=('ScrubObsoleteChanger', )),
    pkg_resources.EntryPoint(
        'debianize', 'silver_platter.debian.debianize',
        attrs=('DebianizeChanger', )),
]


def changer_subcommands() -> List[str]:
    endpoints = pkg_resources.iter_entry_points(
        'silver_platter.debian.changer')
    ret = []
    for ep in BUILTIN_ENTRYPOINTS + list(endpoints):
        ret.append(ep.name)
    return ret


def changer_subcommand(name: str) -> Type[DebianChanger]:
    for ep in BUILTIN_ENTRYPOINTS:
        if ep.name == name:
            return ep.resolve()
    endpoints = pkg_resources.iter_entry_points(
        'silver_platter.debian.changer', name)
    for ep in endpoints:
        return ep.load()
    raise KeyError(name)


def run_mutator(changer_cls, argv=None):
    import json
    import argparse
    import os
    from breezy.workingtree import WorkingTree
    parser = argparse.ArgumentParser()
    changer_cls.setup_parser(parser)
    args = parser.parse_args(argv)
    wt, subpath = WorkingTree.open_containing('.')
    changer = changer_cls.from_args(args)
    try:
        update_changelog_str = os.environ['UPDATE_CHANGELOG'],
    except KeyError:
        update_changelog = None
    else:
        if update_changelog_str == 'leave_changelog':
            update_changelog = False
        elif update_changelog_str == 'update_changelog':
            update_changelog = True
        else:
            # TODO(jelmer): Warn
            update_changelog = None

    try:
        base_metadata_path = os.environ['BASE_METADATA']
    except KeyError:
        existing_proposal = None
    else:
        with open(base_metadata_path, 'r') as f:
            base_metadata = json.load(f)

        class PreviousProposal(MergeProposal):

            def __init__(self, metadata):
                self.metadata = metadata

            def get_description(self):
                return self.metadata.get('description')

            def get_commit_message(self):
                return self.metadata.get('commit-message')

        existing_proposal = PreviousProposal(base_metadata['merge-proposal'])

    mutator_metadata = {}
    try:
        result = changer.make_changes(
            wt, subpath, update_changelog=update_changelog,
            committer=os.environ.get('COMMITTER'),
            base_proposal=existing_proposal,
            reporter=DummyChangerReporter())
    except ChangerError as e:
        result_json = {
            'result-code': e.category,
            'description': e.summary,
        }
    else:
        result_json = {
            'result-code': None,
            'description': result.description,
            'suggested-branch-name': changer.suggest_branch_name(),
            'tags': result.tags,
            'branches': result.branches,
            'value': result.value,
            'mutator': mutator_metadata,
            'merge-proposal': {
                'sufficient': changer.sufficient_for_proposal,
                'commit-message': result.proposed_commit_message,
                'title': result.title,
                'labels': result.labels,
                'description-plain': changer.get_proposal_description(
                    result.mutator, 'plain', existing_proposal),
                'description-markdown': changer.get_proposal_description(
                    result.mutator, 'markdown', existing_proposal),
            }
        }
    json.dump(result_json, sys.stdout, indent=4)
    return 0
